#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import errno
import signal
import logging
import textwrap
import argparse
import traceback

from . import VERSION
from .path import set_path
from .tools import (
    get_tools,
    tool_prepend_prefix,
    python_module_missing,
    tool_required,
    OS_NAMES,
)
from .config import Config
from .environ import normalize_environment
from .logging import line_eraser, setup_logging
from .progress import ProgressManager, Progress
from .profiling import ProfileManager, profile
from .tempfiles import clean_all_temp_files
from .difference import Difference
from .comparators import ComparatorManager
from .external_tools import EXTERNAL_TOOLS, HUGE_TOOLS
from .presenters.html import JQUERY_SYSTEM_LOCATIONS
from .presenters.formats import PresenterManager
from .comparators.utils.compare import compare_root_paths
from .readers import load_diff, load_diff_from_path

logger = logging.getLogger(__name__)


try:
    import tlsh
except ImportError:
    python_module_missing("tlsh")
    tlsh = None

try:
    import argcomplete
except ImportError:
    python_module_missing("argcomplete")
    argcomplete = None


class BooleanAction(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed for BooleanAction")
        super().__init__(option_strings, dest, nargs=0, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, not option_string.startswith("--no"))


def create_parser():
    parser = argparse.ArgumentParser(
        description="Calculate differences between two files or directories",
        add_help=False,
        formatter_class=HelpFormatter,
        fromfile_prefix_chars="@",
    )
    parser.add_argument(
        "path1",
        help="First file or directory to compare.",
    )
    parser.add_argument("path2", help="Second file or directory to compare.")
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Display debug messages",
    )
    parser.add_argument(
        "--pdb",
        action="store_true",
        help="Open the Python pdb debugger in case of crashes",
    )
    parser.add_argument(
        "--status-fd",
        metavar="FD",
        type=int,
        help="Send machine-readable status to file descriptor FD",
    )
    parser.add_argument(
        "--progress",
        "--no-progress",
        action=BooleanAction,
        default=None,
        help="Show an approximate progress bar. Default: yes if "
        "stdin is a tty, otherwise no.",
    )
    parser.add_argument(
        "--no-default-limits",
        action="store_true",
        default=False,
        help="Disable most default output limits and diff calculation limits.",
    )

    group1 = parser.add_argument_group("output types")
    group1.add_argument(
        "--text",
        metavar="OUTPUT_FILE",
        dest="text_output",
        help="Write plain text output to given file (use - for stdout)",
    )
    group1.add_argument(
        "--text-color",
        metavar="WHEN",
        default="auto",
        choices=["never", "auto", "always"],
        help="When to output color diff. WHEN is one of {%(choices)s}. "
        "Default: auto, meaning yes if the output is a terminal, otherwise no.",
    )
    group1.add_argument(
        "--output-empty",
        action="store_true",
        help="If there was no difference, then output an empty "
        "diff for each output type that was specified. In "
        "--text output, an empty file is written.",
    )
    group1.add_argument(
        "--html",
        metavar="OUTPUT_FILE",
        dest="html_output",
        help="Write HTML report to given file (use - for stdout)",
    )
    group1.add_argument(
        "--html-dir",
        metavar="OUTPUT_DIR",
        dest="html_output_directory",
        help="Write multi-file HTML report to given directory",
    )
    group1.add_argument(
        "--css",
        metavar="URL",
        dest="css_url",
        help="Link to an extra CSS for the HTML report",
    )
    group1.add_argument(
        "--jquery",
        metavar="URL",
        dest="jquery_url",
        help="URL link to jQuery, for --html and --html-dir output. "
        "If this is a non-existent relative URL, diffoscope will create a "
        "symlink to a system installation. (Paths searched: %s.) "
        "If not given, --html output will not use JS but --html-dir will "
        'if it can be found; give "disable" to disable JS on all outputs.'
        % ", ".join(JQUERY_SYSTEM_LOCATIONS),
    )
    group1.add_argument(
        "--json",
        metavar="OUTPUT_FILE",
        dest="json_output",
        help="Write JSON text output to given file (use - for stdout)",
    )
    group1.add_argument(
        "--markdown",
        metavar="OUTPUT_FILE",
        dest="markdown_output",
        help="Write Markdown text output to given file (use - for stdout)",
    )
    group1.add_argument(
        "--restructured-text",
        metavar="OUTPUT_FILE",
        dest="restructuredtext_output",
        help="Write RsT text output to given file (use - for stdout)",
    )
    group1.add_argument(
        "--difftool",
        metavar="TOOL",
        dest="difftool",
        help="Compare differences one-by-one using the specified external "
        "command similar to git-difftool(1)",
    )
    group1.add_argument(
        "--profile",
        metavar="OUTPUT_FILE",
        dest="profile_output",
        nargs="?",
        const="-",
        default=None,
        help="Write profiling info to given file (use - for stdout)",
    )
    parser.add_argument(
        "--load-existing-diff",
        metavar="INPUT_FILE",
        action=LoadExistingDiffAction,
        dest="load_existing_diff",
        help='Load existing diff from file. Specify "-" to read a diffoscope diff from stdin.',
    )

    group2 = parser.add_argument_group("output limits")
    # everything marked with default=None below is affected by no-default-limits
    group2.add_argument(
        "--diff-context",
        metavar="LINES",
        type=int,
        help="Lines of unified diff context to show. (default: %d)"
        % Config().diff_context,
        default=Config().diff_context,
    )
    group2.add_argument(
        "--max-text-report-size",
        metavar="BYTES",
        type=int,
        help="Maximum bytes written in --text report. (0 to "
        "disable, default: %d)" % Config().max_text_report_size,
        default=None,
    )
    group2.add_argument(
        "--max-report-size",
        metavar="BYTES",
        type=int,
        help="Maximum bytes of a report in a given format, "
        "across all of its pages. Note that some formats, such "
        "as --html, may be restricted by even smaller limits "
        "such as --max-page-size. (0 to disable, default: %d)"
        % Config().max_report_size,
        default=None,
    ).completer = RangeCompleter(Config().max_report_size)
    group2.add_argument(
        "--max-diff-block-lines",
        metavar="LINES",
        type=int,
        help="Maximum number of lines output per unified-diff "
        "block, across all pages. (0 to disable, default: %d)"
        % Config().max_diff_block_lines,
        default=None,
    ).completer = RangeCompleter(Config().max_diff_block_lines)
    group2.add_argument(
        "--max-page-size",
        metavar="BYTES",
        type=int,
        help="Maximum bytes of the top-level (--html-dir) or sole "
        "(--html) page. (default: %(default)s, remains in effect "
        "even with --no-default-limits)",
        default=Config().max_page_size,
    ).completer = RangeCompleter(Config().max_page_size)
    group2.add_argument(
        "--max-page-diff-block-lines",
        metavar="LINES",
        type=int,
        help="Maximum number of lines output per unified-diff block "
        "on the top-level (--html-dir) or sole (--html) page, before "
        "spilling it into a child page (--html-dir) or skipping the "
        "rest of the diff block. (default: %(default)s, remains in "
        "effect even with --no-default-limits)",
        default=Config().max_page_diff_block_lines,
    ).completer = RangeCompleter(Config().max_page_diff_block_lines)

    group3 = parser.add_argument_group("diff calculation")
    group3.add_argument(
        "--new-file", action="store_true", help="Treat absent files as empty"
    )
    group3.add_argument(
        "--exclude",
        dest="excludes",
        metavar="GLOB_PATTERN",
        action="append",
        default=[],
        help="Exclude files whose names (including any directory part) "
        "match %(metavar)s. Use this option to ignore files based on "
        "their names.",
    )
    group3.add_argument(
        "--exclude-command",
        dest="exclude_commands",
        metavar="REGEX_PATTERN",
        action="append",
        default=[],
        help="Exclude commands that match %(metavar)s. For "
        "example '^readelf.*\\s--debug-dump=info' and '^radare2.*' can take"
        "a long time and differences here are likely secondary differences "
        "caused by something represented elsewhere. Use this option to "
        "disable commands that use a lot of resources.",
    )
    group3.add_argument(
        "--exclude-directory-metadata",
        choices=("auto", "yes", "no", "recursive"),
        help="Exclude directory metadata. Useful if comparing "
        "files whose filesystem-level metadata is not intended "
        "to be distributed to other systems. This is true for "
        "most distributions package builders, but not true "
        "for the output of commands such as `make install`. "
        "Metadata of archive members remain un-excluded "
        'except if "recursive" choice is set. '
        "Use this option to ignore permissions, timestamps, "
        "xattrs etc. Default: 'no' if comparing two "
        "directories, else 'yes'. Note that \"file\" metadata is "
        "actually a property of its containing directory "
        "and is not relevant when distributing the file across "
        "systems.",
    )
    group3.add_argument(
        "--extended-filesystem-attributes",
        "--no-extended-filesystem-attributes",
        action=BooleanAction,
        default=False,
        help="Check potentially-expensive filesystem extended "
        "attributes such as POSIX ACLs, lsattr(1)/chattr(1) "
        "attributes etc. (default: False)",
    )
    group3.add_argument(
        "--diff-mask",
        metavar="REGEX_PATTERN",
        dest="diff_masks",
        action="append",
        default=[],
        help="Replace/unify substrings that match regular expression "
        "%(metavar)s from output strings before applying diff. For example, to "
        "filter out a version number or changed path.",
    )
    group3.add_argument(
        "--fuzzy-threshold",
        type=int,
        help="Threshold for fuzzy-matching "
        "(0 to disable, %(default)s is default, 400 is high fuzziness)",
        default=Config().fuzzy_threshold,
    ).completer = RangeCompleter(400)
    group3.add_argument(
        "--tool-prefix-binutils",
        metavar="PREFIX",
        help="Prefix for binutils program names, e.g. "
        '"aarch64-linux-gnu-" for a foreign-arch binary or "g" '
        "if you're on a non-GNU system.",
    )
    group3.add_argument(
        "--max-diff-input-lines",
        metavar="LINES",
        type=int,
        help="Maximum number of lines fed to diff(1) "
        "(0 to disable, default: %d)" % Config().max_diff_input_lines,
        default=None,
    ).completer = RangeCompleter(Config().max_diff_input_lines)
    group3.add_argument(
        "--max-container-depth",
        metavar="DEPTH",
        type=int,
        help="Maximum depth to recurse into containers. "
        "(Cannot be disabled for security reasons, default: "
        "%(default)s)",
        default=Config().max_container_depth,
    )
    group3.add_argument(
        "--timeout",
        metavar="SECONDS",
        type=int,
        help="Best-effort attempt at a global timeout in seconds. If enabled, "
        "diffoscope will not recurse into any further sub-archives after X "
        "seconds of total execution time. (default: no timeout) [experimental]",
        default=float("inf"),
    )
    group3.add_argument(
        "--max-diff-block-lines-saved",
        metavar="LINES",
        type=int,
        help="Maximum number of lines saved per diff block. "
        "Most users should not need this, unless you run out "
        "of memory. This truncates diff(1) output before emitting "
        "it in a report, and affects all types of output, "
        "including --text and --json. (0 to disable, default: "
        "%(default)s)",
        default=0,
    )
    group3.add_argument(
        "--use-dbgsym",
        metavar="WHEN",
        default="auto",
        choices=("no", "auto", "yes"),
        help="When to automatically use corresponding -dbgsym packages when "
        "comparing .deb files. WHEN is one of {%(choices)s}. Default: auto, "
        "meaning yes if two .changes or .buildinfo files are specified, "
        "otherwise no.",
    )
    group3.add_argument(
        "--force-details",
        default=False,
        action="store_true",
        help="Force recursing into the depths of file formats "
        "even if files have the same content, only really "
        "useful for debugging diffoscope. Default: %(default)s",
    )

    group4 = parser.add_argument_group("information commands")
    group4.add_argument(
        "--help", "-h", action="help", help="Show this help and exit"
    )
    group4.add_argument(
        "--version",
        action="version",
        version="diffoscope %s" % VERSION,
        help="Show program's version number and exit",
    )
    group4.add_argument(
        "--list-tools",
        nargs="?",
        type=str,
        action=ListToolsAction,
        metavar="DISTRO",
        choices=OS_NAMES,
        help="Show external tools required and exit. "
        "DISTRO can be one of {%(choices)s}. "
        "If specified, the output will list packages in that "
        "distribution that satisfy these dependencies.",
    )
    group4.add_argument(
        "--list-debian-substvars",
        action=ListDebianSubstvarsAction,
        help="List packages needed for Debian in 'substvar' format.",
    )
    group4.add_argument(
        "--list-missing-tools",
        nargs="?",
        type=str,
        action=ListMissingToolsAction,
        metavar="DISTRO",
        choices=OS_NAMES,
        help="Show missing external tools and exit. "
        "DISTRO can be one of {%(choices)s}. "
        "If specified, the output will list packages in that "
        "distribution that satisfy these dependencies.",
    )

    if not tlsh:
        parser.epilog = (
            "File renaming detection based on fuzzy-matching is currently "
            'disabled. It can be enabled by installing the "tlsh" module '
            "available from https://github.com/trendmicro/tlsh or in the "
            "python3-tlsh package."
        )
    if argcomplete:
        argcomplete.autocomplete(parser)
    elif "_ARGCOMPLETE" in os.environ:
        logger.error(
            'Argument completion requested but the "argcomplete" module is '
            "not installed. It can be obtained from "
            "https://pypi.python.org/pypi/argcomplete or in the "
            "python3-argcomplete package."
        )
        sys.exit(1)

    def post_parse(parsed_args):
        if parsed_args.load_existing_diff is not None:
            # warn about unusual flags in this mode
            ineffective_flags = [
                f
                for x in group3._group_actions
                if getattr(parsed_args, x.dest) != x.default
                for f in x.option_strings
            ]
            if ineffective_flags:
                logger.warning(
                    "Loading diff instead of calculating it, but "
                    "diff-calculation flags were given; they will be ignored:"
                )
                logger.warning(ineffective_flags)

    return parser, post_parse


class HelpFormatter(argparse.HelpFormatter):
    def format_help(self, *args, **kwargs):
        val = super().format_help(*args, **kwargs)

        # Only append the file formats if --help is passed.
        if not set(sys.argv) & {"--help", "-h"}:
            return val

        def append(title, content, indent=24, max_width=78):
            nonlocal val
            wrapped = textwrap.fill(content, max_width - indent)
            val += "\n{}:\n{}\n".format(
                title, textwrap.indent(wrapped, " " * indent)
            )

        append(
            "file formats supported",
            ComparatorManager().format_descriptions(),
        )

        append("diffoscope homepage", "<https://diffoscope.org/>")

        append(
            "bugs/issues",
            "<https://salsa.debian.org/reproducible-builds/diffoscope/issues>",
            max_width=sys.maxsize,
        )

        return val


class RangeCompleter:
    def __init__(self, start, end=0, divisions=16):
        if end < start:
            tmp = end
            end = start
            start = tmp
        self.choices = range(
            start, end + 1, int((end - start + 1) / divisions)
        )

    def __call__(self, prefix, **kwargs):
        return (str(i) for i in self.choices if str(i).startswith(prefix))


class LoadExistingDiffAction(argparse._StoreAction):
    def __call__(self, parser, *args, **kwargs):
        actions_by_dest = {x.dest: x for x in parser._actions}

        # If we have passed a value for --load_existing_diff we don't require
        # path1 or path2 anymore.
        actions_by_dest["path1"].required = False
        actions_by_dest["path2"].required = False

        # Actually store the value.
        super().__call__(parser, *args, **kwargs)


class ListToolsAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.only_missing = False

    def __call__(self, parser, namespace, os_override, option_string=None):
        # Ensure all comparators are imported so tool_required.all is
        # populated.
        ComparatorManager().reload()

        for k, v in sorted(get_tools(self.only_missing).items()):
            print("%s: %s" % (k, ", ".join(v)))

        sys.exit(0)


class ListMissingToolsAction(ListToolsAction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.only_missing = True


class ListDebianSubstvarsAction(argparse._StoreTrueAction):
    def __call__(self, *args, **kwargs):
        # Attempt to import all comparators so tool_required.all is as
        # populated as possible...
        ComparatorManager().reload()

        # ... however for the generated substvar to be effective/deterministic
        # regardless of the currently installed packages, we special-case some
        # tools (NB. not package names) as their modules may not have been
        # imported by the `ComparatorManager().reload()` call above. (#908072)
        tools = set(
            ("gpg", "rpm2cpio")  # comparators/debian.py  # comparators/rpm.py
        )
        tools.update(tool_required.all)

        packages = set()
        packages_minimal = set()
        for x in tools:
            try:
                pkg = EXTERNAL_TOOLS[x]["debian"]
            except KeyError:
                continue

            packages.add(pkg)
            if x not in HUGE_TOOLS:
                packages_minimal.add(pkg)

        # Exclude "Required" packages
        for x in ("gzip", "tar", "coreutils", "diffutils", "findutils"):
            packages.discard(x)
            packages_minimal.discard(x)

        description = "File formats supported include: {}".format(
            ComparatorManager().format_descriptions(),
        )
        wrapped = "${Newline}".join(textwrap.wrap(description, 79))
        wrapped = wrapped.replace("${Newline}.", "${Newline}")
        print("diffoscope:Description={}".format(wrapped))

        print("diffoscope:Recommends={}".format(", ".join(sorted(packages))))
        print(
            "diffoscope-minimal:Recommends={}".format(
                ", ".join(sorted(packages_minimal))
            )
        )

        sys.exit(0)


def configure(parsed_args):
    for x in (
        "max_report_size",
        "max_text_report_size",
        "max_diff_block_lines",
        "max_diff_block_lines_saved",
        "max_diff_input_lines",
    ):
        # Apply limits affected by "no-default-limits"
        v = getattr(parsed_args, x)
        if v is not None:
            setattr(Config(), x, float("inf") if v == 0 else v)

        elif parsed_args.no_default_limits:
            setattr(Config(), x, float("inf"))

    Config().diff_context = parsed_args.diff_context
    Config().max_page_size = parsed_args.max_page_size
    Config().max_page_diff_block_lines = parsed_args.max_page_diff_block_lines

    Config().difftool = parsed_args.difftool
    Config().new_file = parsed_args.new_file
    Config().use_dbgsym = parsed_args.use_dbgsym
    Config().force_details = parsed_args.force_details
    Config().fuzzy_threshold = parsed_args.fuzzy_threshold
    Config().timeout = parsed_args.timeout
    Config().max_container_depth = parsed_args.max_container_depth

    Config().excludes = parsed_args.excludes
    Config().exclude_commands = parsed_args.exclude_commands
    Config().exclude_directory_metadata = (
        parsed_args.exclude_directory_metadata
    )
    Config().extended_filesystem_attributes = (
        parsed_args.extended_filesystem_attributes
    )
    Config().diff_masks = parsed_args.diff_masks

    Config().compute_visual_diffs = PresenterManager().compute_visual_diffs()

    tool_prepend_prefix(
        parsed_args.tool_prefix_binutils,
        "ar",
        "as",
        "ld",
        "ld.bfd",
        "nm",
        "objcopy",
        "objdump",
        "ranlib",
        "readelf",
        "strip",
    )

    Config().check_constraints()


def run_diffoscope(parsed_args):
    """
    (This should not be considered a stable API suitable for external
    consumption, and the lack of configuration of globals may result in
    unpredictable behaviour.)
    """

    logger.debug("Starting diffoscope %s", VERSION)

    ProfileManager().setup(parsed_args)
    PresenterManager().configure(parsed_args)
    if not tlsh and "PYTEST_CURRENT_TEST" not in os.environ:
        logger.info(
            'Fuzzy-matching is currently disabled as the "tlsh" module is unavailable.'
        )

    configure(parsed_args)
    set_path()
    normalize_environment()

    path1, path2 = parsed_args.path1, parsed_args.path2

    # Should we be loading an existing diff from a file
    if parsed_args.load_existing_diff:
        x = parsed_args.load_existing_diff

        if x == "-":
            logger.debug("Loading diff from stdin")
            difference = load_diff(sys.stdin, "stdin")
        else:
            logger.debug(
                "Loading diff from %s", parsed_args.load_existing_diff
            )
            difference = load_diff_from_path(x)
    else:
        if Config().exclude_directory_metadata in ("auto", None):
            # Default to ignoring metadata directory...
            Config().exclude_directory_metadata = "yes"
            if os.path.isdir(path1) and os.path.isdir(path2):
                # ... except if we passed two directories.
                Config().exclude_directory_metadata = "no"

        logger.debug("Starting comparison")
        with Progress():
            with profile("main", "outputs"):
                difference = compare_root_paths(path1, path2)
        ProgressManager().finish()
    # Generate an empty, null diff to write, saving the exit code first.
    has_differences = bool(difference is not None)
    if difference is None and parsed_args.output_empty:
        difference = Difference(path1, path2)
    with profile("main", "outputs"):
        PresenterManager().output(difference, parsed_args, has_differences)
    return 1 if has_differences else 0


def sigterm_handler(signo, stack_frame):
    clean_all_temp_files()
    os._exit(2)


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    signal.signal(signal.SIGTERM, sigterm_handler)

    try:
        import libarchive  # noqa
    except (ImportError, AttributeError):
        traceback.print_exc()
        python_module_missing("libarchive")
        print(
            "\nMissing or incomplete libarchive module. Try installing your "
            "system's 'libarchive' package.",
            file=sys.stderr,
        )
        sys.exit(2)

    parsed_args = None

    try:
        with profile("main", "parse_args"):
            parser, post_parse = create_parser()
            parsed_args = parser.parse_args(args)

        log_handler = ProgressManager().setup(parsed_args)

        with setup_logging(parsed_args.debug, log_handler) as _:
            post_parse(parsed_args)
            # Call main entry point
            sys.exit(run_diffoscope(parsed_args))

    except BrokenPipeError:
        sys.exit(2)
    except OSError as e:
        if e.errno != errno.ENOSPC:
            raise
        logger.error("No space left on device. Diffoscope exiting.")
        sys.exit(2)
    except KeyboardInterrupt:
        logger.error("Keyboard Interrupt")
        sys.exit(2)
    except Exception:
        sys.stderr.buffer.write(line_eraser())
        traceback.print_exc()
        if parsed_args and parsed_args.pdb:
            import pdb

            pdb.post_mortem()
        sys.exit(2)
    finally:
        # Helps our tests run more predictably - some of them call main()
        # which sets Config() values.
        Config().reset()

        with profile("main", "cleanup"):
            clean_all_temp_files()

        # Print profiling output at the very end
        if parsed_args is not None:
            ProfileManager().finish(parsed_args)


if __name__ == "__main__":
    main()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import logging
import operator

from diffoscope.config import Config

try:
    import tlsh
except ImportError:  # noqa
    tlsh = None

logger = logging.getLogger(__name__)


def perform_fuzzy_matching(members1, members2):
    threshold = Config().fuzzy_threshold

    if tlsh is None or threshold == 0:
        return

    # Create local copies because they will be modified by consumer
    members1 = dict(members1)
    members2 = dict(members2)

    seen = set()
    for name1, (file1, _) in members1.items():
        if file1.is_directory() or not file1.fuzzy_hash:
            continue

        comparisons = []
        for name2, (file2, _) in members2.items():
            if name2 in seen or file2.is_directory() or not file2.fuzzy_hash:
                continue
            comparisons.append(
                (tlsh.diff(file1.fuzzy_hash, file2.fuzzy_hash), name2)
            )

        if not comparisons:
            continue

        comparisons.sort(key=operator.itemgetter(0))
        score, name2 = comparisons[0]

        suffix = "will not compare files"
        if score < threshold:
            seen.add(name2)
            yield name1, name2, score
            suffix = "will compare files"

        logger.debug(
            "Fuzzy matching %s %s (score: %d/400): %s",
            name1,
            name2,
            score,
            suffix,
        )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2021‒2022 Chris Lamb <lamby@debian.org>
# Copyright © 2021 Sergei Trofimovich
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import binascii
import dis
import io
import marshal
import re
import struct
import time
import types

from diffoscope.difference import Difference

from .missing_file import MissingFile
from .utils.file import File

re_memory_address = re.compile(r" at 0x\w+(?=, )")


class PycFile(File):
    DESCRIPTION = "Python .pyc files"
    FILE_TYPE_RE = re.compile(r"^python .*byte-compiled$")

    def compare_details(self, other, source=None):
        if isinstance(other, MissingFile):
            return []

        try:
            return [
                Difference.from_text(
                    describe_pyc(self.path),
                    describe_pyc(other.path),
                    self.path,
                    other.path,
                    source="Python bytecode",
                )
            ]
        except ValueError as exc:
            self.add_comment("Could not decompile bytecode: {}".format(exc))
            return []


def describe_pyc(filename):
    with open(filename, "rb") as f:
        return "\n".join(parse_pyc(f))


def parse_pyc(f):
    magic = f.read(4)
    yield "magic:    {}".format(hexlify(magic))

    f.seek(4, 1)
    moddate = f.read(4)
    modtime = time.asctime(time.gmtime(struct.unpack("<L", moddate)[0]))
    yield "moddate:  {} ({} UTC)".format(hexlify(moddate), modtime)

    filesz = f.read(4)
    filesz = struct.unpack("<L", filesz)
    yield f"files sz: {filesz[0]}"

    code = marshal.load(f)
    yield from show_code(code)


def show_code(code, indent=""):
    yield f"{indent}code"

    indent += "   "

    for x in ("argcount", "nlocals", "stacksize", "flags"):
        yield "{}{: <10}: {!r}".format(indent, x, getattr(code, f"co_{x}"))

    yield from show_hex("code", code.co_code, indent=indent)
    s = io.StringIO()
    dis.disassemble(code, file=s)
    for x in s.getvalue().splitlines():
        yield "{}{}".format(indent, re_memory_address.sub("", x))

    yield f"{indent}consts"
    for const in code.co_consts:
        if isinstance(const, types.CodeType):
            yield from show_code(const, f"{indent}   ")
        else:
            yield f"   {indent}{const!r}"

    for x in (
        "names",
        "varnames",
        "freevars",
        "cellvars",
        "filename",
        "name",
        "firstlineno",
    ):
        yield "{}{: <10} {!r}".format(indent, x, getattr(code, f"co_{x}"))

    yield from show_hex("lnotab", code.co_lnotab, indent=indent)


def show_hex(label, val, indent):
    val = hexlify(val)

    if len(val) < 60:
        yield f"{indent}{label} {val}"
        return

    yield f"{indent}{label}"
    for i in range(0, len(val), 60):
        yield "{}   {}".format(indent, val[i : i + 60])


def hexlify(val):
    return "0x{}".format(binascii.hexlify(val).decode("utf-8"))



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import sys
import shutil
import os.path
import zipfile

from diffoscope.config import Config
from diffoscope.tools import tool_required
from diffoscope.difference import Difference
from diffoscope.exc import ContainerExtractionError, RequiredToolNotFound
from diffoscope.tempfiles import get_named_temporary_file

from .utils.file import File
from .directory import Directory
from .utils.archive import Archive, ArchiveMember
from .utils.command import Command


class Zipinfo(Command):
    # zipinfo returns with an exit code of 1 or 2 when reading
    # Mozilla-optimized or Java "jmod" ZIPs as they have non-standard headers
    # which are safe to ignore.
    VALID_RETURNCODES = {0, 1, 2}

    re_strip_path = re.compile(r"^(warning|error) \[[^\]]+\]:\s+(.*)$")

    @tool_required("zipinfo")
    def cmdline(self):
        return ["zipinfo", self.path]

    def filter(self, line):
        # we don't care about the archive file path
        if line.startswith(b"Archive:"):
            return b""

        # Strip paths from errors and warnings
        # eg: "warning [/full/path]: 472 extra bytes at beginning or within zipfile"
        m = self.re_strip_path.match(line.decode("utf-8"))
        if m is not None:
            return "{}: {}\n".format(m.group(1), m.group(2)).encode("utf-8")

        return line


class ZipinfoVerbose(Zipinfo):
    @tool_required("zipinfo")
    def cmdline(self):
        return ["zipinfo", "-v", self.path]


class Zipnote(Command):
    # zipnote returns with an exit code of 3 for invalid archives
    VALID_RETURNCODES = {0, 3}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.flag = False

    @tool_required("zipnote")
    def cmdline(self):
        path = self.path
        if not path.endswith(".zip"):
            path = get_named_temporary_file(suffix=".zip").name
            shutil.copy(self.path, path)
        return ["zipnote", path]

    def filter(self, line):
        """
        Example output from zipnote(1):

            @ foo
            hello
            @ (comment above this line)
            @ (zip file comment below this line)
            goodbye
        """

        if line == b"@ (zip file comment below this line)\n":
            self.flag = True
            return b"Zip file comment: "

        if line == b"@ (comment above this line)\n":
            self.flag = False
            return b"\n\n"  # spacer

        if line.startswith(b"@ "):
            filename = line[2:-1].decode()
            self.flag = True
            return f"Filename: {filename}\nComment: ".encode()

        return line[:-1] if self.flag else b""


class BsdtarVerbose(Command):
    @tool_required("bsdtar")
    def cmdline(self):
        return ["bsdtar", "-tvf", self.path]


def zipinfo_differences(file, other):
    """
    Run all our zipinfo variants.
    """

    for x in (Zipinfo, ZipinfoVerbose, BsdtarVerbose):
        result = Difference.from_operation(x, file.path, other.path)
        # We only return the 'best' one
        if result is not None:
            return [result]

    return []


class ZipDirectory(Directory, ArchiveMember):
    def __init__(self, archive, member_name):
        ArchiveMember.__init__(self, archive, member_name)

    def compare(self, other, source=None):
        return None

    def has_same_content_as(self, other):
        return False

    def is_directory(self):
        return True

    def get_member_names(self):
        raise ValueError("Zip archives are compared as a whole.")  # noqa

    def get_member(self, member_name):
        raise ValueError("Zip archives are compared as a whole.")  # noqa


class ZipContainer(Archive):
    def open_archive(self):
        return zipfile.ZipFile(self.source.path, "r")

    def close_archive(self):
        self.archive.close()

    def get_member_names(self):
        return self.archive.namelist()

    def extract(self, member_name, dest_dir):
        # We don't really want to crash if the filename in the zip archive
        # can't be encoded using the filesystem encoding. So let's replace
        # any weird character so we can get to the bytes.
        targetpath = os.path.join(
            dest_dir, os.path.basename(member_name)
        ).encode(sys.getfilesystemencoding(), errors="replace")

        try:
            with self.archive.open(member_name) as source, open(
                targetpath, "wb"
            ) as target:
                shutil.copyfileobj(source, target)
            return targetpath.decode(sys.getfilesystemencoding())
        except RuntimeError as e:
            # Handle encrypted files see line 1292 of zipfile.py
            is_encrypted = self.archive.getinfo(member_name).flag_bits & 0x1
            if is_encrypted:
                raise ContainerExtractionError(member_name, e)
            raise

    def get_member(self, member_name):
        zipinfo = self.archive.getinfo(member_name)
        if zipinfo.filename[-1] == "/":
            return ZipDirectory(self, member_name)
        return ArchiveMember(self, member_name)


class ZipFileBase(File):
    def compare(self, other, source=None):
        x = super().compare(other, source)

        if x is None:
            return None

        # If we have two or more differences, then we have observed differences
        # within a nested file. If not, there is likely some difference in
        # the metadata that zipinfo cannot discover, so we manually fallback to
        # a binary diff.
        if len(x.details) >= 2:
            return x

        x.add_comment(
            "Archive contents identical but files differ, possibly due "
            "to different compression levels. Falling back to binary "
            "comparison."
        )

        details = self.compare_bytes(other, source=source)
        if details is not None:
            x.add_details([details])

        return x


class ZipFile(ZipFileBase):
    DESCRIPTION = "ZIP archives"
    CONTAINER_CLASSES = [ZipContainer]
    FILE_TYPE_RE = re.compile(
        r"^((?:iOS App )?Zip archive|Java archive|EPUB document|OpenDocument (Text|Spreadsheet|Presentation|Drawing|Formula|Template|Text Template)|Google Chrome extension)\b"
    )

    def compare_details(self, other, source=None):
        differences = []
        if Config().exclude_directory_metadata != "recursive":
            differences.extend(zipinfo_differences(self, other))

        try:
            differences.append(
                Difference.from_operation(Zipnote, self.path, other.path)
            )
        except RequiredToolNotFound:  # noqa
            pass

        return differences


class MozillaZipContainer(ZipContainer):
    def open_archive(self):
        # This is gross: Monkeypatch zipfile._EndRecData to work with
        # Mozilla-optimized ZIPs
        _orig_EndRecData = zipfile._EndRecData

        def _EndRecData(fh):
            endrec = _orig_EndRecData(fh)
            if endrec:
                endrec[zipfile._ECD_LOCATION] = (
                    endrec[zipfile._ECD_OFFSET] + endrec[zipfile._ECD_SIZE]
                )
            return endrec

        zipfile._EndRecData = _EndRecData
        result = super(MozillaZipContainer, self).open_archive()
        zipfile._EndRecData = _orig_EndRecData
        return result


class MozillaZipFile(ZipFile):
    DESCRIPTION = "Mozilla-optimized .ZIP archives"
    CONTAINER_CLASSES = [MozillaZipContainer]

    @classmethod
    def recognizes(cls, file):
        # Mozilla-optimized ZIPs start with a 32-bit little endian integer
        # indicating the amount of data to preload, followed by the ZIP
        # central directory (with a PK\x01\x02 signature)
        return file.file_header[4:8] == b"PK\x01\x02"


class JmodJavaModule(ZipFile):
    DESCRIPTION = "Java .jmod modules"
    FILE_TYPE_RE = re.compile(r"^(Zip archive data|Java jmod module)")

    @classmethod
    def recognizes(cls, file):
        # Not all versions of file(1) support the detection of these
        # modules yet so we perform our own manual check for a "JM" prefix.
        return file.file_header[:2] == b"JM"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import logging
import subprocess

from diffoscope.tools import tool_required
from diffoscope.config import Config
from diffoscope.difference import Difference

from .image import pixel_difference, flicker_difference, same_size
from .utils.file import File
from .utils.command import Command, our_check_output

logger = logging.getLogger(__name__)


class Gifbuild(Command):
    RE_FILTERS = (
        re.compile(r"^# GIF information from"),
        re.compile(r"^# End of .* dump"),
    )

    @tool_required("gifbuild")
    def cmdline(self):
        return ["gifbuild", "-d", self.path]

    def filter(self, line):
        if any(x.match(line.decode("utf-8")) for x in self.RE_FILTERS):
            return b""
        return line


@tool_required("identify")
def is_image_static(image):
    try:
        return (
            our_check_output(("identify", "-format", "%n", image.path)) == b"1"
        )
    except subprocess.CalledProcessError:
        return False


def can_compose_gif_images(image1, image2):
    return (
        same_size(image1, image2)
        and is_image_static(image1)
        and is_image_static(image2)
    )


class GifFile(File):
    DESCRIPTION = "GIF image files"
    FILE_TYPE_RE = re.compile(r"^GIF image data\b")

    def compare_details(self, other, source=None):
        gifbuild_diff = Difference.from_operation(
            Gifbuild, self.path, other.path, source="gifbuild"
        )

        differences = [gifbuild_diff]

        if (
            gifbuild_diff is not None
            and Config().compute_visual_diffs
            and can_compose_gif_images(self, other)
        ):
            try:
                logger.debug(
                    "Generating visual difference for %s and %s",
                    self.path,
                    other.path,
                )
                content_diff = Difference(
                    self.path, other.path, source="Image content"
                )
                content_diff.add_visuals(
                    [
                        pixel_difference(self.path, other.path),
                        flicker_difference(self.path, other.path),
                    ]
                )
                differences.append(content_diff)
            except subprocess.CalledProcessError:  # noqa
                pass

        return differences



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import logging
import subprocess
import collections
import hashlib

from diffoscope.exc import OutputParsingError
from diffoscope.tools import get_tool_name, tool_required
from diffoscope.config import Config
from diffoscope.tempfiles import get_named_temporary_file
from diffoscope.difference import Difference
from diffoscope.utils import format_class

from .deb import DebFile, get_build_id_map
from .decompile import DecompilableContainer
from .utils.file import File
from .utils.command import Command, our_check_output

DEBUG_SECTION_GROUPS = (
    "rawline",
    "info",
    "abbrev",
    "pubnames",
    "aranges",
    "macro",
    "frames",
    "loc",
    "ranges",
    "pubtypes",
    "trace_info",
    "trace_abbrev",
    "trace_aranges",
    "gdb_index",
)

logger = logging.getLogger(__name__)


class Readelf(Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # we don't care about the name of the archive
        self._path_re = re.compile(
            r"\b%s/\b" % re.escape(os.path.dirname(self.path))
        )
        self._archive_re = re.compile(r"^File: %s\(" % re.escape(self.path))

    @tool_required("readelf")
    def cmdline(self):
        return (
            [get_tool_name("readelf"), "--wide"]
            + self.readelf_options()
            + [self.path]
        )

    def readelf_options(self):
        return []  # noqa

    def filter(self, line):
        try:
            val = line.decode("utf-8")
        except UnicodeDecodeError:
            return line

        # we don't care about the name of the archive
        val = self._archive_re.sub("File: lib.a(", val)

        # the full path can appear in the output, we need to remove it
        val = self._path_re.sub("/", val)

        return val.encode("utf-8")

    @staticmethod
    def should_skip_section(section_name, section_type):
        return False


class ReadelfFileHeader(Readelf):
    def readelf_options(self):
        return ["--file-header"]


class ReadelfProgramHeader(Readelf):
    def readelf_options(self):
        return ["--program-header"]


class ReadelfSections(Readelf):
    def readelf_options(self):
        return ["--sections"]


class ReadelfSymbols(Readelf):
    def readelf_options(self):
        return ["--symbols"]

    @staticmethod
    def should_skip_section(section_name, section_type):
        return section_type in {"DYNSYM", "SYMTAB"}


class ReadelfRelocs(Readelf):
    def readelf_options(self):
        return ["--relocs"]

    @staticmethod
    def should_skip_section(section_name, section_type):
        return section_type in {"REL", "RELA"}


class ReadelfDynamic(Readelf):
    def readelf_options(self):
        return ["--dynamic"]

    @staticmethod
    def should_skip_section(section_name, section_type):
        return section_type == "DYNAMIC"


class ReadelfNotes(Readelf):
    def readelf_options(self):
        return ["--notes"]

    @staticmethod
    def should_skip_section(section_name, section_type):
        return section_type == "NOTE"


class RedaelfVersionInfo(Readelf):
    def readelf_options(self):
        return ["--version-info"]

    @staticmethod
    def should_skip_section(section_name, section_type):
        return section_type in {"VERDEF", "VERSYM", "VERNEED"}


class ReadelfDebugDump(Readelf):
    def readelf_options(self):
        return [f"--debug-dump={self._debug_section_group}"]


READELF_DEBUG_DUMP_COMMANDS = [
    type(
        f"ReadelfDebugDump_{x}",
        (ReadelfDebugDump,),
        {"_debug_section_group": x},
    )
    for x in DEBUG_SECTION_GROUPS
]


class ReadElfSection(Readelf):
    @staticmethod
    def base_options():
        if not hasattr(ReadElfSection, "_base_options"):
            output = our_check_output(
                [get_tool_name("readelf"), "--help"],
                stderr=subprocess.DEVNULL,
            ).decode("us-ascii", errors="replace")

            ReadElfSection._base_options = []
            for x in ("--decompress",):
                if x in output:
                    ReadElfSection._base_options.append(x)
        return ReadElfSection._base_options

    def __init__(self, path, section_name, *args, **kwargs):
        self._path = path
        self._section_name = section_name
        super().__init__(path, *args, **kwargs)

    @property
    def section_name(self):
        return self._section_name

    def readelf_options(self):
        return ReadElfSection.base_options() + [
            "--hex-dump={}".format(self.section_name)
        ]


class ReadelfStringSection(ReadElfSection):
    def readelf_options(self):
        return ReadElfSection.base_options() + [
            "--string-dump={}".format(self.section_name)
        ]


class ObjdumpSection(Command):
    # Remove offsets before instructions (see issue #17)
    RE_INSTRUCTION_OFFSET = re.compile(rb"^\s*[0-9a-f]+:")

    def __init__(self, path, section_name, *args, **kwargs):
        self._path = path
        self._path_bin = path.encode("utf-8")
        self._section_name = section_name
        super().__init__(path, *args, **kwargs)

    def objdump_options(self):
        return []

    @tool_required("objdump")
    def cmdline(self):
        return (
            [get_tool_name("objdump")]
            + self.objdump_options()
            + ["--section={}".format(self._section_name), self.path]
        )

    def filter(self, line):
        # Remove the filename from the output
        if line.startswith(self._path_bin + b":"):
            return b""
        if line.startswith(b"In archive"):
            return b""

        line = ObjdumpSection.RE_INSTRUCTION_OFFSET.sub(b"", line)

        # objdump (from approximately 2.35.50.20201125-1 onwards) returns "ret"
        # instead of "retq"; lets normalise them to "retq" for now.
        if line == b"\tret    \n":
            return b"\tretq   \n"

        return line


class ObjdumpDisassembleSection(ObjdumpSection):
    RE_SYMBOL_COMMENT = re.compile(
        rb"^( +[0-9a-f]+:[^#]+)# [0-9a-f]+ <[^>]+>$"
    )

    def objdump_options(self):
        # With '--line-numbers' we get the source filename and line within the
        # disassembled instructions.
        # objdump can get the debugging information from the elf or from the
        # stripped symbols file specified in the .gnu_debuglink section
        return [
            "--line-numbers",
            "--disassemble",
            "--demangle",
            "--reloc",
            "--no-show-raw-insn",
        ]

    def filter(self, line):
        # Apply this filter first as super() removes offsets
        line = ObjdumpDisassembleSection.RE_SYMBOL_COMMENT.sub(r"\1", line)
        return super().filter(line)


class ObjdumpDisassembleSectionNoLineNumbers(ObjdumpDisassembleSection):
    def objdump_options(self):
        return ["--disassemble", "--demangle", "--no-show-raw-insn"]


READELF_COMMANDS = (
    ReadelfFileHeader,
    ReadelfProgramHeader,
    ReadelfSections,
    ReadelfSymbols,
    ReadelfRelocs,
    ReadelfDynamic,
    ReadelfNotes,
    RedaelfVersionInfo,
)


def _should_skip_section(name, type):
    for x in READELF_COMMANDS:
        if x.should_skip_section(name, type):
            logger.debug("Skipping section %s, covered by %s", name, x)
            return True
    if name.startswith(".debug") or name.startswith(".zdebug"):
        # section .debug_str looks much nicer with `readelf --string-dump`
        # the rest is handled by READELF_DEBUG_DUMP_COMMANDS
        return not name.endswith("_str")
    return False


class ElfSection(File):
    def __init__(self, elf_container, member_name):
        super().__init__(container=elf_container)
        self._name = member_name

    @property
    def name(self):
        return self._name

    @property
    def progress_name(self):
        return "{} [{}]".format(
            self.container.source.progress_name, super().progress_name
        )

    @property
    def path(self):
        return self.container.source.path

    def cleanup(self):
        pass

    def is_directory(self):
        return False

    def is_symlink(self):
        return False

    def is_device(self):
        return False

    def has_same_content_as(self, other):
        # Always force diff of the section
        return False

    @property
    def fuzzy_hash(self):
        return None

    @classmethod
    def recognizes(cls, file):
        # No file should be recognized as an elf section
        return False

    def compare(self, other, source=None):
        return Difference.from_operation(
            ReadElfSection, self.path, other.path, operation_args=[self._name]
        )


class ElfCodeSection(ElfSection):
    def compare(self, other, source=None):
        # Disassemble with line numbers, but if the command is excluded or
        # fails, fallback to disassembly. If that is also excluded or failing,
        # only then fallback to a hexdump.
        diff = None
        try:
            diff, excluded = Difference.from_operation_exc(
                ObjdumpDisassembleSection,
                self.path,
                other.path,
                operation_args=[self._name],
            )
        except subprocess.CalledProcessError as e:
            # eg. When failing to disassemble a different architecture.
            logger.error(e)

        if diff and not excluded:
            return diff

        try:
            diff, excluded = Difference.from_operation_exc(
                ObjdumpDisassembleSectionNoLineNumbers,
                self.path,
                other.path,
                operation_args=[self._name],
            )
        except subprocess.CalledProcessError as e:
            logger.error(e)

        if diff and not excluded:
            return diff

        return super().compare(other, source)


class ElfStringSection(ElfSection):
    def compare(self, other, source=None):
        return Difference.from_operation(
            ReadelfStringSection,
            self.path,
            other.path,
            operation_args=[self._name],
        )


@tool_required("readelf")
def get_build_id(path):
    try:
        output = our_check_output(
            [get_tool_name("readelf"), "--notes", path],
            stderr=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError as e:
        logger.debug("Unable to get Build ID for %s: %s", path, e)
        return None

    m = re.search(
        r"^\s+Build ID: ([0-9a-f]+)$",
        output.decode("utf-8"),
        flags=re.MULTILINE,
    )
    if not m:
        return None

    return m.group(1)


@tool_required("readelf")
def get_debug_link(path):
    try:
        output = our_check_output(
            [get_tool_name("readelf"), "--string-dump=.gnu_debuglink", path],
            stderr=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError as e:
        logger.debug("Unable to get Build Id for %s: %s", path, e)
        return None

    m = re.search(
        r"^\s+\[\s+0\]\s+(\S+)$",
        output.decode("utf-8", errors="replace"),
        flags=re.MULTILINE,
    )
    if not m:
        return None

    return m.group(1)


class ElfContainer(DecompilableContainer):
    auto_diff_metadata = False

    SECTION_FLAG_MAPPING = {
        "X": ElfCodeSection,
        "S": ElfStringSection,
        "_": ElfSection,
    }

    @tool_required("readelf")
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logger.debug("Creating ElfContainer for %s", self.source.path)

        cmd = [
            get_tool_name("readelf"),
            "--wide",
            "--section-headers",
            self.source.path,
        ]
        output = our_check_output(cmd, shell=False, stderr=subprocess.DEVNULL)
        has_debug_symbols = False
        has_build_id = False

        try:
            output = output.decode("utf-8").split("\n")
            if output[1].startswith("File:"):
                output = output[2:]
            output = output[5:]

            # Entries of readelf --section-headers have the following columns:
            # [Nr]  Name  Type  Address  Off  Size  ES  Flg  Lk  Inf  Al
            self._sections = collections.OrderedDict()
            for line in output:
                if line.startswith("Key to Flags"):
                    break

                # Strip number column because there may be spaces in the brakets
                line = line.split("]", 1)[1].split()
                name, type, flags = line[0], line[1], line[6] + "_"

                if name.startswith(".debug") or name.startswith(".zdebug"):
                    has_debug_symbols = True

                if name == ".note.gnu.build-id" and type == "NOTE":
                    has_build_id = True

                if _should_skip_section(name, type):
                    continue

                # Use first match, with last option being '_' as fallback
                elf_class = [
                    ElfContainer.SECTION_FLAG_MAPPING[x]
                    for x in flags
                    if x in ElfContainer.SECTION_FLAG_MAPPING
                ][0]

                logger.debug(
                    "Adding section %s (%s) as %s",
                    name,
                    type,
                    format_class(
                        elf_class, strip="diffoscope.comparators.elf."
                    ),
                )
                self._sections[name] = elf_class(self, name)

        except Exception as e:
            command = " ".join(cmd)
            logger.debug(
                "OutputParsingError in %s from `%s` output - %s:%s",
                self.__class__.__name__,
                command,
                e.__class__.__name__,
                e,
            )
            raise OutputParsingError(command, self)

        if not has_debug_symbols:
            self._install_debug_symbols()

        if has_build_id:
            try:
                self._verify_build_id()
            except Exception:
                # It is fine to skip the verification of the build_id
                pass

    @tool_required("objcopy")
    def _install_debug_symbols(self):
        if Config().use_dbgsym == "no":
            return

        # Figure out if we are in a Debian package first
        try:
            deb = (
                self.source.container.source.container.source.container.source
            )
        except AttributeError:
            return

        # It needs to be a .deb and we need access a to a -dbgsym package in
        # the same .changes, directory or archive
        if not isinstance(deb, DebFile) or not deb.container:
            return

        # If the .deb in question is the top-level of the source we have passed
        # a .deb directly to diffoscope (versus finding one specified in a
        # .changes or .buildinfo file). In this case, don't automatically
        # search for a -dbgsym file unless the user specified
        # `Config().use_dbgsym`.
        if (
            not hasattr(deb.container.source, "container")
            and Config().use_dbgsym != "yes"
        ):
            return

        # Retrieve the Build ID for the ELF file we are examining
        build_id = get_build_id(self.source.path)
        debuglink = get_debug_link(self.source.path)
        if not build_id or not debuglink:
            return

        logger.debug(
            "Looking for a dbgsym package for Build Id %s (debuglink: %s)",
            build_id,
            debuglink,
        )

        # Build a map of Build-Ids if it doesn't exist yet
        if not hasattr(deb.container, "dbgsym_build_id_map"):
            deb.container.dbgsym_build_id_map = get_build_id_map(deb.container)

        if build_id not in deb.container.dbgsym_build_id_map:
            logger.debug(
                "Unable to find a matching debug package for Build Id %s",
                build_id,
            )
            return

        dbgsym_package = deb.container.dbgsym_build_id_map[build_id]
        debug_file_path = "./usr/lib/debug/.build-id/{0}/{1}.debug".format(
            build_id[:2], build_id[2:]
        )
        debug_file = (
            dbgsym_package.as_container.data_tar.as_container.lookup_file(
                debug_file_path
            )
        )
        if not debug_file:
            logger.debug(
                "Unable to find the matching debug file %s in %s",
                debug_file_path,
                dbgsym_package,
            )
            return

        # Create a .debug directory and link the debug symbols there with the
        # right name
        dest_path = os.path.join(
            os.path.dirname(self.source.path),
            ".debug",
            os.path.basename(debuglink),
        )
        os.makedirs(os.path.dirname(dest_path), exist_ok=True)

        def objcopy(*args):
            our_check_output(
                (get_tool_name("objcopy"),) + args,
                shell=False,
                stderr=subprocess.DEVNULL,
            )

        # If #812089 was fixed, we would just do os.link(debug_file.path,
        # dest_path) but for now, we need to do more complicated things…
        # 1. Use objcopy to create a file with only the original .gnu_debuglink
        # section as we will have to update it to get the CRC right.
        debuglink_path = get_named_temporary_file(
            prefix="{}.debuglink.".format(self.source.path)
        ).name

        objcopy(
            "--only-section=.gnu_debuglink", self.source.path, debuglink_path
        )

        # 2. Monkey-patch the ElfSection object created for the .gnu_debuglink
        # to change the path to point to this new file
        section = self._sections[".gnu_debuglink"]

        class MonkeyPatchedElfSection(section.__class__):
            @property
            def path(self):
                return debuglink_path

        section.__class__ = MonkeyPatchedElfSection

        # 3. Create a file with the debug symbols in uncompressed form
        objcopy("--decompress-debug-sections", debug_file.path, dest_path)

        # 4. Update the .gnu_debuglink to this new file so we get the CRC right
        objcopy("--remove-section=.gnu_debuglink", self.source.path)
        objcopy("--add-gnu-debuglink={}".format(dest_path), self.source.path)

        logger.debug("Installed debug symbols at %s", dest_path)

    def _verify_build_id(self):
        """
        Verify whether the NT_GNU_BUILD_ID field contains a sha1 checksum
        that matches the binary. (#260)
        """

        with open(self.source.path, "rb") as f:
            blob = f.read()

        # Magic value: length=0x14, owner length=3, owner='GNU', followed by the sha1 checksum
        m = re.search(
            b"\x14\x00\x00\x00\x03\x00\x00\x00\x47\x4e\x55\x00.{20}", blob
        )
        build_id = blob[m.end() - 20 : m.end()].hex()
        blob_with_reset_build_id = (
            blob[: m.end() - 20] + b"\x00" * 20 + blob[m.end() :]
        )

        if hashlib.sha1(blob_with_reset_build_id).hexdigest() != build_id:
            self.source.add_comment(
                f"File has been modified after NT_GNU_BUILD_ID has been applied."
            )
            logger.debug(
                f"{self.source.path} has been modified after NT_GNU_BUILD_ID has been applied"
            )
            logger.debug(
                "Expected value: %s, current value: %s",
                hashlib.sha1(blob_with_reset_build_id).hexdigest(),
                build_id,
            )
        return

    def get_member_names(self):
        decompiled_members = super().get_member_names()
        return list(decompiled_members) + list(self._sections.keys())

    def get_member(self, member_name):
        try:
            return self._sections[member_name]
        except KeyError:
            # Raised when the member name is not one of ours, which means
            # it was part of the decompiler's output (aka super)
            return super().get_member(member_name)


class Strings(Command):
    re_debug_line = re.compile(r"^\s?\w{38,40}\.debug\n$")
    re_gcc_line = re.compile(r"^.{1,4}GCC: \([^\)]+\)")

    @tool_required("strings")
    def cmdline(self):
        return ("strings", "--all", "--bytes=8", self.path)

    def filter(self, line):
        val = line.decode("utf-8")

        # Don't include specific ".debug"-like lines in the output, as it
        # invariably a duplicate of the debug ID that exists better in the
        # readelf(1) differences for this file.
        #
        # For example:
        #
        #   b'0684311d738d2555027e737d5846e7478561fa.debug\n'
        #   b'·Xu94356212e19a62aadc4db9c3c93076ac997706.debug\n'
        #
        if self.re_debug_line.match(val):
            return b""

        # Don't include short "GCC" lines that differs on a single prefix byte
        # either. These are distracting, not very useful and are simply the
        # strings(1) command's idea of the build ID, which is displayed
        # elsewhere in the diff.
        #
        # For example:
        #
        #   │ -GCC: (Debian 10.2.1-6) 10.2.1 20210110
        #   │ +QGCC: (Debian 10.2.1-6) 10.2.1 20210110
        #
        if self.re_gcc_line.match(val):
            return line

        return line


class ElfFile(File):
    DESCRIPTION = "ELF binaries"
    CONTAINER_CLASSES = [ElfContainer]
    FILE_TYPE_RE = re.compile(r"^ELF ")

    def compare_details(self, other, source=None):
        differences = [
            Difference.from_operation(
                x, self.path, other.path, ignore_returncodes={1}
            )
            for x in list(READELF_COMMANDS) + READELF_DEBUG_DUMP_COMMANDS
        ]

        difference = Difference.from_operation(Strings, self.path, other.path)
        if difference:
            difference.check_for_ordering_differences()
        differences.append(difference)

        return differences



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2016, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import stat

from .utils.file import File


class FilesystemFile(File):
    def __init__(self, path, container=None):
        super().__init__(container=container)
        self._name = path

    @property
    def path(self):
        return self._name

    def is_directory(self):
        return not os.path.islink(self._name) and os.path.isdir(self._name)

    def is_symlink(self):
        return os.path.islink(self._name)

    def is_device(self):
        mode = os.lstat(self._name).st_mode
        return stat.S_ISCHR(mode) or stat.S_ISBLK(mode)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
# Copyright © 2018 Paul Wise <pabs@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import json
import signal
import logging

from .logging import line_eraser
from .tools import python_module_missing


logger = logging.getLogger(__name__)

try:
    import progressbar
except ImportError:
    python_module_missing("progressbar")
    progressbar = None


class ProgressLoggingHandler(logging.StreamHandler):
    def __init__(self, progressbar):
        self.progressbar = progressbar
        super().__init__()

    def emit(self, record):
        try:
            super().emit(record)
            if not getattr(self.progressbar.bar, "finished", False):
                self.progressbar.bar.update()
        except Exception:
            # Wrap otherwise tests fail due to test_progress.py call main()
            # several times. This mirrors the super() implementation.
            self.handleError(record)


class ProgressManager:
    _singleton = {}

    def __init__(self):
        self.__dict__ = self._singleton

        if not self._singleton:
            self.reset()

    def reset(self):
        self.stack = []
        self.observers = []

    def setup(self, parsed_args):
        def show_progressbar():
            # Show progress bar if user explicitly asked for it
            if parsed_args.progress:
                return True

            # ... otherwise show it if STDOUT is a tty and we are not debugging
            if parsed_args.progress is None:
                return sys.stdout.isatty() and not parsed_args.debug

            return False

        log_handler = None
        if show_progressbar():
            if progressbar is not None:
                bar = ProgressBar()
                self.register(bar)
                log_handler = ProgressLoggingHandler(bar)
            else:
                # User asked for bar, so show them an error

                if parsed_args.progress:
                    logging.warning(
                        'Progress bar was requested but the "progressbar" module is unavailable.'
                    )

        if parsed_args.status_fd:
            self.register(StatusFD(os.fdopen(parsed_args.status_fd, "w")))

        return log_handler

    def push(self, progress):
        assert not self.stack or self.stack[-1].is_active()
        self.stack.append(progress)

    def pop(self, progress):
        x = self.stack.pop()
        assert x is progress
        if self.stack:
            self.stack[-1].child_done(x.total)

    def register(self, observer):
        logger.debug("Registering %s as a progress observer", observer)
        self.observers.append(observer)

    def update(self, msg):
        if self.stack:
            cur_estimates = None
            for progress in reversed(self.stack):
                cur_estimates = progress.estimates(cur_estimates)
            current, total = cur_estimates
        else:
            current, total = 0, 1

        for x in self.observers:
            x.notify(current, total, msg)

    def finish(self):
        for x in self.observers:
            x.finish()


class Progress:
    def __init__(self, total=None):
        self.done = []
        self.current_steps = None
        self.current_child_steps_done = None
        if total:
            self.total = total
        else:
            self.total = 1
            self.begin_step(1)

    def __enter__(self):
        ProgressManager().push(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.maybe_end()
        ProgressManager().pop(self)

    def estimates(self, cur_child_estimate=None):
        own_done = sum(pair[0] for pair in self.done)
        children_done = sum(pair[1] for pair in self.done)
        all_done = own_done + children_done

        if self.current_steps:
            if self.current_child_steps_done or cur_child_estimate:
                # Something is in-progress, the calculation is slightly more
                # complex
                cur_child_done, cur_child_total = cur_child_estimate or (0, 0)
                own_done += self.current_steps
                all_done += (
                    self.current_steps
                    + self.current_child_steps_done
                    + cur_child_done
                )
                # Cost of what we expect will have been done, once the current
                # in-progress step plus all of its children, have completed
                expected_all_done = all_done + (
                    cur_child_total - cur_child_done
                )
                assert own_done  # non-zero
                return (
                    all_done,
                    int(float(self.total) / own_done * expected_all_done),
                )
        else:
            # nothing in progress
            assert not cur_child_estimate

        if not own_done:
            assert not children_done
            return 0, self.total

        # weigh self.total by (all_done/own_done)
        return all_done, int(float(self.total) / own_done * all_done)

    def is_active(self):
        return self.current_steps is not None

    def maybe_end(self, msg=""):
        if self.is_active():
            self.done += [(self.current_steps, self.current_child_steps_done)]
            self.current_steps = None
            self.current_child_steps_done = None
            ProgressManager().update(msg)

    def begin_step(self, step, msg=""):
        assert step is not None
        self.maybe_end(msg)
        self.current_steps = step
        self.current_child_steps_done = 0

    def child_done(self, total):
        self.current_child_steps_done += total


class ProgressBar:
    def __init__(self):
        try:
            from progressbar.widgets import WidgetBase
        except ImportError:
            # Fallback to the older Debian version
            from progressbar import Widget as WidgetBase

        self.msg = ""

        class Message(WidgetBase):
            def update(self, pbar, _observer=self):
                msg = _observer.msg
                width = 25

                if len(msg) <= width:
                    return msg.rjust(width)

                # Print the last `width` characters with an ellipsis.
                return "…{}".format(msg[-width + 1 :])

        class OurProgressBar(progressbar.ProgressBar):
            def __init__(self, *args, **kwargs):
                # Remove after https://github.com/niltonvolpato/python-progressbar/pull/57 is fixed.
                kwargs.setdefault("fd", sys.stderr)
                super().__init__(*args, **kwargs)
                # Terminal handling after parent init since that sets self.fd
                self.erase_to_eol = line_eraser(self.fd)

            def _need_update(self):
                return True

            def erase_line(self):
                if self.erase_to_eol:
                    self.fd.buffer.write(self.erase_to_eol)
                    self.fd.flush()

            def finish(self):
                self.finished = True
                self.update(self.maxval)
                # Clear the progress bar after completion
                self.erase_line()
                if self.signal_set:
                    signal.signal(signal.SIGWINCH, signal.SIG_DFL)

        self.bar = OurProgressBar(
            widgets=(
                " ",
                progressbar.Bar(),
                "  ",
                progressbar.Percentage(),
                "  ",
                Message(),
                "  ",
                progressbar.ETA(),
                " ",
            )
        )
        self.bar.start()

    def notify(self, current, total, msg):
        self.msg = msg

        self.bar.maxval = total
        self.bar.currval = current
        self.bar.update()

    def finish(self):
        self.bar.finish()


class StatusFD:
    def __init__(self, fileobj):
        self.fileobj = fileobj

    def notify(self, current, total, msg):
        print(
            json.dumps({"msg": msg, "total": total, "current": current}),
            file=self.fileobj,
        )

    def finish(self):
        pass
